#!/usr/bin/env python
# Copyright:
# ----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as authorized
# by a licensing agreement from ARM Limited.
#      (C) COPYRIGHT 2019-2020 ARM Limited, ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorized copies and
# copies may only be made to the extent permitted by a licensing agreement
# from ARM Limited.
# ----------------------------------------------------------------------------
#

import argparse
import logging
import os
import re
import sys


if __name__ == "__main__":
    # This can be run as a plugin or a standalone script - if it's a standalone
    # script, set up sys.path so the `config_system` module can be imported.
    sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "bob", "config_system"))
    logger = logging.getLogger(os.path.basename(__file__))
    logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
else:
    logger = logging.getLogger(__name__)


import config_system
import config_system.log_handlers


def log_handler_supports_colors(handler):
    if isinstance(handler, logging.StreamHandler) and handler.stream.isatty():
        return True
    elif isinstance(handler, config_system.log_handlers.ErrorCounterHandler):
        return True
    return False


def all_log_handlers_support_colors(logger):
    for handler in logger.handlers:
        if not log_handler_supports_colors(handler):
            return False

    if not logger.parent:
        return True
    return all_log_handlers_support_colors(logger.parent)


if all_log_handlers_support_colors(logger):
    format_params = {
        "boldred": "\033[1;31m",
        "red": "\033[;31m",
        "reset": "\033[0m",
        "yellow": "\033[;33m",
    }
else:
    format_params = {
        "boldred": "",
        "red": "",
        "reset": "",
        "yellow": "",
    }

format_params["width"] = 62
format_params["blank"] = ""
format_params["top"] = "-"
format_params["side"] = format_params["yellow"] + "|" + format_params["reset"]


class BannerStyle:
    top = \
"""{yellow}+{top:{top}^{width}}+{reset}
{side}{blank:^{width}}{side}"""
    bottom = \
"""{yellow}+{top:{top}^{width}}+{reset}"""
    issue = \
"""{side}{red}{message:^{width}}{side}
{side}{blank:^{width}}{side}
{options}
{side}{blank:^{width}}{side}"""

    option = "{side}{option:^{width}}{side}"


class NormalStyle:
    top = ""
    bottom = ""
    issue = "{red}{message}{reset}\n{options}"
    option = " * {option}"


def format_config_issue(message, options, style):
    """Generate a string containing the warning/error in `message` followed by
    the list of options triggering it.
    """
    options = [style.option.format(option=i, **format_params) for i in options]
    return style.issue.format(
            message=message,
            options="\n".join(options),
            **format_params)

def format_config_warning(message, options, style):
    """Format warnings for config options that select a special config option,
    e.g. WARN_ON or DEPRECATED. The function will generate a string containing
    the header text in `message` followed by the list of options with their
    corresponding "warning" text, if specified.
    """
    warning_msgs = []
    for opt in options:
        warning_msgs.append(opt)
        warning = config_system.get_warning(opt)
        if warning:
            warning_msgs[-1] += ": " + warning

    options = [style.option.format(option=warn_msg, **format_params) for warn_msg in warning_msgs]
    return style.issue.format(
            message=message,
            options="\n".join(options),
            **format_params)

def log_message(message, log):
    """Output a message, prefixing each line with the current logging level.
    """
    for line in message.splitlines():
        log(line)

def get_bad_android_options():
    """Get options which are deprecated or not supported for Android
    userspace. Using GCC on Android is deprecated.
    """
    options = []

    android = config_system.get_config_bool("ANDROID")
    target_gcc = config_system.get_config_bool("TARGET_TOOLCHAIN_GNU")
    host_gcc = config_system.get_config_bool("HOST_TOOLCHAIN_GNU")

    if android and (target_gcc or host_gcc):
        options.append("ANDROID")
        if target_gcc:
            options.append("TARGET_TOOLCHAIN_GNU")
        if host_gcc:
            options.append("HOST_TOOLCHAIN_GNU")

    return options

def get_bad_android_kernel_options():
    """Get options which are deprecated or not supported for Android
    kernel modules. Using GCC on Android is deprecated.
    """
    options = []

    android = config_system.get_config_bool("ANDROID")
    api_level = config_system.get_config_int("ANDROID_API_LEVEL")
    kernel_cc = config_system.get_config_string("KERNEL_CC")
    build_kernel = config_system.get_config_bool("BUILD_KERNEL_MODULES")

    if kernel_cc == "":
        # When empty we use the kernels default compiler, gcc
        kernel_gcc = True
    else:
        gcc_re = r".*gcc(-[\d]+)?$"
        kernel_gcc = re.match(gcc_re, kernel_cc)

    if android and api_level > 280 and build_kernel and kernel_gcc:
        options.append("ANDROID")
        options.append("BUILD_KERNEL_MODULES")
        options.append("ANDROID_API_LEVEL={}".format(api_level))
        options.append("KERNEL_CC={}".format(kernel_cc))

    return options


def check_config(allow_banner):
    """Check for enabled options which either select DEPRECATED, depend on
    UNRELEASED or are a combination not supported on Android.
    Allow printing with two different styles of message, depending
    on the value of `allow_banner`.
    """

    android_options = get_bad_android_options()
    android_kernel_options = get_bad_android_kernel_options()
    deprecated_options = config_system.get_options_selecting("DEPRECATED")
    unreleased_options = config_system.get_options_depending_on("UNRELEASED")
    unsupported_options = config_system.get_options_selecting("UNSUPPORTED")
    warnon_options = config_system.get_options_selecting("WARN_ON")

    if not (deprecated_options or unreleased_options or unsupported_options or
            android_options or android_kernel_options):
        return

    style = NormalStyle
    if allow_banner:
        if android_options or android_kernel_options:
            style = BannerStyle
        elif deprecated_options:
            style = BannerStyle
        elif (unreleased_options and
              config_system.get_config_string("RELEASE_NAME") != "not_released"):
            style = BannerStyle
        elif unsupported_options:
            style = BannerStyle

    warnings, errors = [], []
    if android_options:
        errors.append(format_config_issue("Compiling with GCC on Android is unsupported:",
                                          android_options, style))
    if android_kernel_options:
        errors.append(format_config_issue("Compiling kernel modules with GCC is unsupported:",
                                          android_kernel_options, style))
    if deprecated_options:
        warnings.append(format_config_warning("The following enabled options are deprecated:",
                                              deprecated_options, style))
    if unreleased_options:
        warnings.append(format_config_warning("The following enabled options are not released:",
                                              unreleased_options, style))
    if unsupported_options:
        errors.append(format_config_warning("The following enabled options are not supported:",
                                            unsupported_options, style))
    if warnon_options:
        warnings.append(format_config_warning("The following enabled options have warnings associated with them:",
                                              warnon_options, style))

    if len(warnings) > 0 or len(errors) > 0:
        log_warn = logger.warning
        log_err = lambda line: logger.error('  ' + line)
        log_top = log_warn if len(warnings) > 0 else log_err
        log_bottom = log_err if len(errors) > 0 else log_warn

        log_message(style.top.format(**format_params), log_top)
        log_message("\n".join(warnings), log_warn)
        log_message("\n".join(errors), log_err)
        log_message(style.bottom.format(**format_params), log_bottom)

def plugin_exec():
    # When run as a config system plugin, allow the large "banner"-style
    # messages, because this is during the config step, when the user can still
    # change the configuration accordingly.
    check_config(True)


def main():
    ap = argparse.ArgumentParser("Check for deprecated and unreleased configuration")
    ap.add_argument("-c", "--config", required=True,
                    help="Path to the configuration file (*.config)")
    ap.add_argument("-d", "--database", required=True,
                    help="Path to the configuration database (Mconfig)")
    ap.add_argument("--ignore-missing", dest="ignore_missing", action="store_true", default=False,
                    help="Ignore missing database files included with 'source'")
    args = ap.parse_args()

    config_system.read_config(args.database, args.config, args.ignore_missing)

    # When run from main(), this is during the build - so only allow "small"
    # messages, because this is just a reminder; it's too late to change the
    # configuration.
    check_config(False)


if __name__ == "__main__":
    main()
